#!/usr/bin/env python
"""@package CreateFlowtable

@brief Command line driver program to create multiple RHESSys flowtables,
one flowtable for each worldfile.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2015, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import sys
import os
import argparse

from rhessysworkflows.command.exceptions import *
from rhessysworkflows.command.flowtable import FlowtableMultiple

if __name__ == "__main__":
    
    # Handle command line options
    parser = argparse.ArgumentParser(description='Create RHESSys flowtables for multiple worldfiles using GRASS GIS data and createflowpaths utility')
    parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                        help='The configuration file. Must define section "GRASS" and option "GISBASE"')
    parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                        help='The directory to which metadata, intermediate, and final files should be saved')
    parser.add_argument('--routeRoads', dest='routeRoads', required=False, action='store_true',
                        help='Tell createflowpaths to route flow from roads to the nearest stream pixel (requires roads_rast to be defined in metadata)')
    parser.add_argument('--routeRoofs', dest='routeRoofs', required=False, action='store_true',
                        help='Tell createflowpaths to route flow from roof tops based on roof top connectivity to nearest impervious surface (requires roof_connectivity_rast and impervious_rast to be defined in metadata)')
    parser.add_argument('-f', '--force', dest='force', action='store_true',
                        help='Run createflowpaths even if DEM x resolution does not match y resolution')
    parser.add_argument('--ignoreBurnedDEM', dest='ignoreBurnedDEM', action='store_true', required=False,
                        help='Ignore stream burned DEM, if present. Default DEM raster will be used for all operations. If not specified and if stream burned raster is present, stream burned DEM will be used for generating the flow table.')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='Print detailed information about what the program is doing')
    args = parser.parse_args()

    configFile = None
    if args.configfile:
        configFile = args.configfile
        
    command = FlowtableMultiple(args.projectDir, configFile)
    
    exitCode = os.EX_OK
    try: 
        command.run(routeRoads=args.routeRoads, 
                    routeRoofs=args.routeRoofs,
                    ignoreBurnedDEM=args.ignoreBurnedDEM,
                    force=args.force,
                    verbose=args.verbose)
    except CommandException as e:
        print(str(e))
        exitCode = os.EX_DATAERR
    
    sys.exit(exitCode)


#!/usr/bin/env python
"""@package PatchToCumulativeMap

@brief Tool for making cummulative maps of patch-scale RHESSys output variables.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2014, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'

2. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset


Post conditions
---------------
None
   
@todo Add ability to read in two files, accumulate data, and write normalized map (value / max(value))   

"""
import os, sys, shutil, tempfile, re
import subprocess, shlex
import time, random
import argparse
import operator

import numpy as np
import statsmodels.api as sm
import matplotlib.pyplot as plt

from ecohydrolib.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from ecohydrolib.grasslib import *

from rhessysworkflows.rhessys import RHESSysOutput

PATCH_DAILY_RE = re.compile('^(.+_patch.daily)$')
VARIABLE_EXPR_RE = re.compile(r'\b([a-zA-z][a-zA-Z0-9_\.]+)\b') # variable names can have "." in them
RECLASS_MAP_TMP = "patchtomovietmp_%d" % (random.randint(100000, 999999),)

# Handle command line options
parser = argparse.ArgumentParser(description='Generate cummulative map of patch-scale RHESSys output variables')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file.')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-d', '--rhessysOutFile', required=True, nargs='+',
                    help='Directory containing RHESSys patch output, specificically patch daily output stored in a file whose name ends with "_patch.daily".')
parser.add_argument('--mask', required=False, default=None,
                    help='Name of raster to use as a mask')
parser.add_argument('--overlay', required=False, default=None, nargs='*',
                    help='Name of raster map to use as overlay')
parser.add_argument('--overlayLegend', required=False, default=False,
                    help='Display legend for overlay layer. If more than one overlay is present, will only draw legend for the first.')
parser.add_argument('-o', '--outputDir', required=True,
                    help='Directory to which map should be output')
parser.add_argument('-f', '--outputFile', required=True, nargs='+',
                    help='Name of file to store map in; .png extension will be added.  If file exists it will be overwritten.')
parser.add_argument('-c', '--cdfOutputfile', required=True,
                    help='Name of file to write CDF output to')
parser.add_argument('--patchMap', required=False, default='patch',
                    help='Name of patch map')
parser.add_argument('-v', '--outputVariable', required=True,
                    help='Name of RHESSys variable to be mapped.  Can be an expression such as "trans_sat + trans_unsat"')
parser.add_argument('-t' ,'--mapTitle', required=False,
                    help='Text to use for title.  If not supplied, variable name will be used')
args = parser.parse_args()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile)

if len(args.rhessysOutFile) != len(args.outputFile):
    sys.exit("Number of data files %d does not match number of output filenames %d" % \
             (len(args.rhessysOutFile), len(args.outputFile)) )

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))

patchDailyFilepaths = []
for outfile in args.rhessysOutFile:
    if not os.path.isfile(outfile) or not os.access(outfile, os.R_OK):
        sys.exit("Unable to read RHESSys output file %s" % (outfile,))
    patchDailyFilepaths.append( os.path.abspath(outfile) ) 

outputFilePaths = []
for outfile in args.outputFile:
    if not os.path.isdir(args.outputDir) or not os.access(args.outputDir, os.W_OK):
        sys.exit("Unable to write to output directory %s" % (args.outputDir,) )
    outputDir = os.path.abspath(args.outputDir)
    outputFile = "%s.png" % (outfile,)
    outputFilePaths.append( os.path.join(outputDir, outputFile) )

cdfFilepath = os.path.join(outputDir, args.cdfOutputfile)

# Determine output variables
variables = ['patchID']
m = VARIABLE_EXPR_RE.findall(args.outputVariable)
if m:
    variables += m
else:
    sys.exit("No output variables specified")

if args.mapTitle:
    title = args.mapTitle
else:
    title = args.outputVariable

# 1. Get tmp folder for temprarily storing rules
tmpDir = tempfile.mkdtemp()
print("Temp dir: %s" % (tmpDir,) )
reclassRule = os.path.join(tmpDir, 'reclass.rule')
colorTable = os.path.join(tmpDir, 'color.rule')
# Create our own color ramp  
with open(colorTable, 'w') as colorsOut:
    colorsOut.write('0 96:0:0\n')
    colorsOut.write('0.001 96:32:0\n')
    colorsOut.write('0.01 128:64:32\n')
    colorsOut.write('0.1 160:96:64\n')
    colorsOut.write('0.2 192:128:96\n')
    colorsOut.write('0.4 224:160:128\n')
    colorsOut.write('0.8 255:192:160\n')
    colorsOut.write('0.9 255:224:192\n')
    colorsOut.write('0.95 255:255:224\n')
    colorsOut.write('1 255:255:255\n')

# 2. Initialize GRASS
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

# Set mask (if present)
if args.mask:
    result = grassLib.script.run_command('r.mask',
                                         input=args.mask,
                                         maskcats="1",
                                         flags="o")
    if result != 0:
        sys.exit("Failed to set mask using layer %s" % (args.mask,) )
    # Zoom to map extent
    result = grassLib.script.run_command('g.region',
                                         rast='MASK',
                                         zoom='MASK')
    if result != 0:
        sys.exit("Failed to set region to layer %s" % \
                 (args.mask,) )
        
# Set environment variables for GRASS PNG output driver
os.environ['GRASS_RENDER_IMMEDIATE'] = 'FALSE'
os.environ['GRASS_TRUECOLOR'] = 'TRUE'
os.environ['GRASS_WIDTH'] = '960'
os.environ['GRASS_HEIGHT'] = '720'

# 3. For each rhessys output file ...
variablesList = []
for (i, patchDailyFilepath) in enumerate(patchDailyFilepaths):
    print("\nReading RHESSys output %s from %s  (this may take a while)...\n" \
          % (os.path.basename(patchDailyFilepath), os.path.dirname(patchDailyFilepath)) )
    f = open(patchDailyFilepath)
    data = RHESSysOutput.readColumnsFromPatchDailyFile(f, variables)
    f.close()
    if len(data) < 1:
        sys.exit("No data found for variable in RHESSys output file '%s'" % \
                 (patchDailyFilepath,) )
    
    # For each day
    patchIDs = None
    variable = None
    for (i, key) in enumerate(data):
        dataForDate = data[key]
        expr = VARIABLE_EXPR_RE.sub(r'np.array(dataForDate["\1"])', args.outputVariable)
        if variable is None:
            variable = eval(expr)
        else:
            variable += eval(expr)
        if patchIDs is None:
            patchIDs = [ int(f) for f in dataForDate['patchID'] ]
    
    print("Sum of cumulative %s = %.2f" % (args.outputVariable, variable.sum()) )
    print("Mean of cumulative %s = %.2f" % (args.outputVariable, variable.mean()) )
    print("Max of cumulative %s = %.2f" % (args.outputVariable, variable.max()) )
    print("Min of cumulative %s = %.2f\n" % (args.outputVariable, variable.min()) )
    variablesList.append(variable)
        
# 4. Normalize values to maximum  
max_val = 0
for var in variablesList:
    max_val = max( max_val, np.max(var) )
    
print("\nMax cumulative %s = %.2f\n" % (args.outputVariable, max_val) )
max_idx = None
max_patchID = None
# Find the max value
for var in variablesList:
    max_idx = np.where(var == max_val)
    # Could be multiple indices
    if len(max_idx[0]) == 1 and max_idx[0] >= 0: 
        max_patchID = patchIDs[max_idx[0]]
        break;
if max_patchID:
    print("\nPatchID of max value: %d" % (max_patchID,) )
 
normalizedVariables = []
for var in variablesList:
    normalizedVariables.append( var / max_val )
        
# Write normalized maps for each input file
for (i, variable) in enumerate(normalizedVariables):
    outputFilePath = outputFilePaths[i]
    # 5. Write reclass rule to temp file
    reclass = open(reclassRule, 'w') 
    for (j, var) in enumerate(variable):
        reclass.write("%d:%d:%f:%f\n" % (patchIDs[j], patchIDs[j], var, var) )
    reclass.close()
        
    # 6. Generate map for variable
    result = grassLib.script.run_command('r.recode', 
                                         input=args.patchMap, 
                                         output=RECLASS_MAP_TMP,
                                         rules=reclassRule,
                                         overwrite=True)
    if result != 0:
        sys.exit("Failed to create reclass map while rendering image %s" % (outputFilePath,) )
    
    # Set color table
    result = grassLib.script.run_command('r.colors', 
                                         map=RECLASS_MAP_TMP,
                                         rules=colorTable)
    if result != 0:
        sys.exit("Failed to modify color map while rendering image %s" % (outputFilePath,) )
    
    # 7. Render map with annotations to PNG image
    # Start a new PNG driver
    os.environ['GRASS_PNGFILE'] = outputFilePath
    result = grassLib.script.run_command('d.mon', start='PNG')
    if result != 0:
        sys.exit("Failed to start PNG driver while rendering image %s" % \
                 (outputFilePath,) )
    
    # Render image
    result = grassLib.script.run_command('d.rast',
                                         map=RECLASS_MAP_TMP)
    if result != 0:
        sys.exit("Failed to render map %s while rendering image %s" % \
                 (RECLASS_MAP_TMP, outputFilePath) )
       
    # Draw overlays 
    if args.overlay:
        for (i, overlay) in enumerate(args.overlay):
            result = grassLib.script.run_command('d.rast',
                                                 map=overlay,
                                                 flags='o')
            if result != 0:
                sys.exit("Failed to draw overlay map %s while rendering image %s" % \
                         (overlay, outputFilePath) )
            if i == 0 and args.overlayLegend:
                # Add legend for overlay
                result = grassLib.script.run_command('d.legend',
                                                 map=overlay,
                                                 at='15,50,10,15')
                if result != 0:
                    sys.exit("Failed to add legend for overlay map %s while rendering image %s" % \
                             (args.overlay, outputFilePath) )
                # Write name of overlay layer
                result = grassLib.script.run_command('d.text',
                                                 text=overlay,
                                                 size=2.5,
                                                 color='black',
                                                 at='12, 12',
                                                 align='cc')
                if result != 0:
                    sys.exit("Failed to add overlay annotation to map while rendering image %s" % \
                             (outputFilePath,) )
            
    # Add annotations
    result = grassLib.script.run_command('d.legend',
                                         map=RECLASS_MAP_TMP,
                                         at='63,88,85,95')
    if result != 0:
        sys.exit("Failed to add legend to map while rendering image %s" % \
                 (outputFilePath,) )
    
    # Set title
    result = grassLib.script.run_command('d.text',
                                         text=title,
                                         size=5,
                                         color='black',
                                         at='73,98',
                                         align='cc')
    if result != 0:
        sys.exit("Failed to add title to map while rendering image %s" % \
                 (outputFilePath,) )
    
    # Write map imaage to file
    result = grassLib.script.run_command('d.mon', stop='PNG')
    if result != 0:
        sys.exit("Error occured when closing PNG driver for image %s" % \
                 (outputFilePath,) )

# Plot CDF
fig = plt.figure(figsize=(8, 6), dpi=80, tight_layout=True)
ax = fig.add_subplot(111)

min_val = 0.01
x = np.linspace(min_val, max_val, num=len(variablesList[0]))

data_plt = []
for var in variablesList:
    tmp_ecdf = sm.distributions.ECDF(var)
    (tmp_plot, ) = ax.plot(x, tmp_ecdf(x))
    data_plt.append(tmp_plot)
ax.set_xlim(min_val, max_val)
ax.set_xscale('log')
ax.set_xlabel(title) 
fig.suptitle("Cumulative distribution - %s" % (title,), y=0.99)
ax.legend( data_plt, args.outputFile, 'upper left', fontsize='x-small', ncol=1, frameon=False)
    
cdfFilepath_png = "%s.png" % (cdfFilepath,)
cdfFilepath_pdf = "%s.pdf" % (cdfFilepath,)
plt.savefig(cdfFilepath_png)
plt.savefig(cdfFilepath_pdf)

# Cleanup
shutil.rmtree(tmpDir)
result = grassLib.script.run_command('g.remove', rast=RECLASS_MAP_TMP)
if result != 0:
    sys.exit("Failed to remove temporary map %s" % (RECLASS_MAP_TMP,) )



#!/usr/bin/env python
"""@package PatchToMap

@brief Tool for mapping patch-scale RHESSys yearly output variables.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2014, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'

2. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset


Post conditions
---------------
None

"""
import os, sys, shutil, tempfile, re
import subprocess, shlex
import time, random
import argparse
import operator
import math

import numpy as np
import statsmodels.api as sm
import matplotlib
import matplotlib.pyplot as plt

from ecohydrolib.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from ecohydrolib.grasslib import *

from rhessysworkflows.rhessys import RHESSysOutput



INT_RESCALE = 100
VARIABLE_EXPR_RE = re.compile(r'\b([a-zA-z][a-zA-Z0-9_\.]+)\b') # variable names can have "." in them
RANDOM = random.randint(100000, 999999)
RECLASS_MAP_TMP = "patchzonalstats_cover_{0}".format(RANDOM)
STATS_MAP_TMP = "patchzonalstats_output_{0}".format(RANDOM)

methods = ['average', 'mode', 'median', 'avedev', 'stddev', 'variance',
           'skewness', 'kurtosis', 'min', 'max', 'sum']

# Handle command line options
parser = argparse.ArgumentParser(description='Generate cummulative map of patch-scale RHESSys output variables')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file.')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-d', '--rhessysOutFile', required=True, nargs='+',
                    help='Directory containing RHESSys patch output, ' +\
                         'specificically patch yearly output in a file ending in "_patch.yearly".')
parser.add_argument('--mask', required=False, default=None,
                    help='Name of raster to use as a mask')
parser.add_argument('--patchMap', required=False, default='patch',
                    help='Name of patch map')
parser.add_argument('-y', '--year', required=False, type=int,
                    help='Year for which statistics should be generated')
parser.add_argument('-v', '--outputVariable', required=True,
                    help='Name of RHESSys variable to be mapped.  Can be an expression such as "trans_sat + trans_unsat"')
parser.add_argument('-n', '--mapNames', required=True, nargs='+',
                    help='Names of maps to be created.')
parser.add_argument('--mapcolorstyle', required=False, default='grey1.0',
                    help='Color map style to pass to r.colors, used for zonal stats map.')
args = parser.parse_args()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile)

if not len(args.rhessysOutFile) == len(args.mapNames):
    sys.exit("Number of RHESSys output files must match the number of map names")

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))

patchFilepaths = []
for outfile in args.rhessysOutFile:
    if not os.path.isfile(outfile) or not os.access(outfile, os.R_OK):
        sys.exit("Unable to read RHESSys output file %s" % (outfile,))
    patchFilepaths.append( os.path.abspath(outfile) ) 

# Determine output variables
variables = ['patchID']
m = VARIABLE_EXPR_RE.findall(args.outputVariable)
if m:
    variables += m
else:
    sys.exit("No output variables specified")

# 1. Get tmp folder for temprarily storing rules
tmpDir = tempfile.mkdtemp()
print("Temp dir: %s" % (tmpDir,) )
reclassRule = os.path.join(tmpDir, 'reclass.rule')

# 2. Initialize GRASS
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

# Set mask (if present)
if args.mask:
    result = grassLib.script.run_command('r.mask',
                                         input=args.mask,
                                         maskcats="1",
                                         flags="o")
    if result != 0:
        sys.exit("Failed to set mask using layer %s" % (args.mask,) )
    # Zoom to map extent
    result = grassLib.script.run_command('g.region',
                                         rast='MASK',
                                         zoom='MASK')
    if result != 0:
        sys.exit("Failed to set region to layer %s" % \
                 (args.mask,) )

# 3. For each rhessys output file read variable(s) of interest...
variablesList = []
for (i, patchFilepath) in enumerate(patchFilepaths):
    print("\nReading RHESSys output %s from %s  (this may take a while)...\n" \
          % (os.path.basename(patchFilepath), os.path.dirname(patchFilepath)) )
    data = np.genfromtxt(patchFilepath, names=True)
    if len(data) < 1:
        sys.exit("No data found for variable in RHESSys output file '%s'" % \
                 (patchFilepath,) )
    
    if args.year:
        data = data[data['year'] == float(args.year)]
    
    patchIDs = [ int(p) for p in data['patchID'] ]
    expr = VARIABLE_EXPR_RE.sub(r'data["\1"]', args.outputVariable)
    variablesList.append(eval(expr))
        
# 4. Write maps for each input file
for (i, variable) in enumerate(variablesList):
    reclass_map = args.mapNames[i]
    # Write reclass rule to temp file
    reclass = open(reclassRule, 'w') 
    for (j, var) in enumerate(variable):
        reclass.write("%d:%d:%f:%f\n" % (patchIDs[j], patchIDs[j], var, var) )
    reclass.close()
        
    # Generate map for variable
    print("\nMapping variable: {0} for input {1} to map named {2}...".format(args.outputVariable, 
                                                                              patchFilepaths[i],
                                                                              reclass_map))
    result = grassLib.script.run_command('r.recode', 
                                         input=args.patchMap, 
                                         output=reclass_map,
                                         rules=reclassRule,
                                         flags='d',
                                         overwrite=True)
    if result != 0:
        sys.exit("Failed to create reclass map {0}".format(reclass_map) )
        
    # Set color table
    if args.mapcolorstyle:
        result = grassLib.script.run_command('r.colors', 
                                             map=reclass_map,
                                             color=args.mapcolorstyle)
        if result != 0:
            sys.exit("Failed to modify color map")

# Cleanup
shutil.rmtree(tmpDir)




"""@package rhessysworkflows.command.giconverter

@brief RHESSysWorkflows commands related to parameterizing green infrastructure.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2016, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os
import sys
import subprocess
import shlex
import importlib
import tempfile

from rhessysworkflows.command.base import GrassCommand
from rhessysworkflows.command.exceptions import MetadataException
from rhessysworkflows.command.exceptions import RunException

from rhessysworkflows.rhessys import RHESSysPaths
from rhessysworkflows.metadata import RHESSysMetadata

from rhessysworkflows.ginotebook import DEFAULT_HOSTNAME, DEFAULT_API_ROOT
from rhessysworkflows.ginotebook import GINotebook


class GIConverter(GrassCommand):

    def __init__(self, projectDir, configFile=None, outfp=sys.stdout):
        """ Construct a GIConverter command.
        Arguments:
        projectDir -- string    The path to the project directory
        configFile -- string    The path to an EcohydroLib configuration file
        outfp -- file-like object    Where output should be written to

        """
        super(GIConverter, self).__init__(projectDir, configFile, outfp)
        self.param_const = self.param_db = None
        self.paths = None

    def checkMetadata(self, *args, **kwargs):
        """ Check to make sure the project directory has the necessary metadata to run this command.
        """
        super(GIConverter, self).checkMetadata(args, kwargs)

        # Check for necessary information in metadata
        if not 'dem_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a DEM raster" % (self.context.projectDir,))
        if not 'soil_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a soil raster in a GRASS mapset" % (self.context.projectDir,))
        if not 'landcover_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a landcover raster" % (self.context.projectDir,))
        if not 'landuse_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a land use raster" % (self.context.projectDir,))
        if not 'stratum_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a stratum raster" % (self.context.projectDir,))
        if not 'patch_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a patch raster in a GRASS mapset" % (self.context.projectDir,))
        if not 'roads_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a roads raster" % (self.context.projectDir,))
        if not 'slope_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a slope raster" % (self.context.projectDir,))
        if not 'streams_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a stream raster" % (self.context.projectDir,))
        if not 'zero_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a zero raster" % (self.context.projectDir,))

        if not 'dem_srs' in self.studyArea:
            raise MetadataException("Metadata in project directory %s does not contain a DEM spatial reference system" % (self.context.projectDir,))
        if not 'dem_res_x' in self.studyArea:
            raise MetadataException("Metadata in project directory %s does not contain a DEM x resolution" % (self.context.projectDir,))
        if not 'dem_res_y' in self.studyArea:
            raise MetadataException("Metadata in project directory %s does not contain a DEM y resolution" % (self.context.projectDir,))

        if not 'grass_dbase' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS Dbase" % (self.context.projectDir,))
        if not 'grass_location' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS location" % (self.context.projectDir,))
        if not 'grass_mapset' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS mapset" % (self.context.projectDir,))
        if not 'rhessys_dir' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a RHESSys directory" % (self.self.context.projectDir,))
        if not 'cf_bin' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a createflowpaths executable" % (self.context.projectDir,))
        if not 'lairead_bin' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a lairead executable" % (self.context.projectDir,))
        if not 'allometric_table' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain an allometric table" % (self.context.projectDir,))
        if not 'template_template' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a template_template" % (self.context.projectDir,))
        if not 'climate_stations' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a list of climate stations" % (self.context.projectDir,))
        if not 'paramdb_dir' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a ParamDB directory" % (self.context.projectDir,))
        if not 'paramdb' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a ParamDB" % (self.context.projectDir,))

    def run(self, *args, **kwargs):
        """ Create flow tables for multiple worldfiles

        Arguments:
        scenario_id -- int    ID of the GI Notebook scenario whose GI instances are to be parameterized.
        auth_token -- string    Authorization token to use for authenticating to the GI Notebook.
        host -- string    Hostname of GI Notebook server. Default: None.
        api_root -- string    The root of the API URL to use. Default: None.
        use_HTTPS -- boolean    Use HTTPS for communication with the GI Notebook.
        force -- boolean        Force overwrite of existing scenario output. Default: False.
        verbose -- boolean    Produce verbose output. Default: False.
        """
        scenario_id = kwargs.get('scenario_id')
        if scenario_id is None:
            raise RunException('Scenario ID was not specified.')
        auth_token = kwargs.get('auth_token')
        if auth_token is None:
            raise RunException('Authorization token was not specified.')
        host = kwargs.get('host', DEFAULT_HOSTNAME)
        api_root = kwargs.get('api_path', DEFAULT_API_ROOT)
        use_HTTPS = kwargs.get('use_HTTPS', False)
        force = kwargs.get('force', False)
        verbose = kwargs.get('verbose', False)

        self.checkMetadata()
        self.param_const, self.param_db = self._init_paramdb()
        self.paths = RHESSysPaths(self.context.projectDir, self.metadata['rhessys_dir'])

        gi_scenario_base = 'gi_scenario'
        gi_scenario_data = "{0}.geojson".format(gi_scenario_base)
        scenario_geojson_path = os.path.join(self.context.projectDir, gi_scenario_data)
        gi_scenario_soils_base = "{0}_wsoils".format(gi_scenario_base)
        gi_scenario_soils = "{0}.geojson".format(gi_scenario_soils_base)
        scenario_soils_geojson_path = os.path.join(self.context.projectDir, gi_scenario_soils)
        gi_scenario_landuse_base = "{0}_landuse".format(gi_scenario_base)
        gi_scenario_landuse = "{0}.geojson".format(gi_scenario_landuse_base)
        scenario_landuse_geojson_path = os.path.join(self.context.projectDir, gi_scenario_landuse)
        gi_scenario_data_key = 'gi_scenario_data'
        gi_scenario_soils_data_key = "{0}_soils".format(gi_scenario_data_key)
        gi_scenario_landuse_data_key = "{0}_landuse".format(gi_scenario_data_key)
        if gi_scenario_data_key in self.metadata:
            if verbose:
                self.outfp.write('Existing GI scenario found.\n')
            if force:
                if verbose:
                    self.outfp.write('Force option specified, overwriting existing GI scenario.\n')
                if os.path.exists(scenario_geojson_path):
                    os.unlink(scenario_geojson_path)
                if os.path.exists(scenario_soils_geojson_path):
                    os.unlink(scenario_soils_geojson_path)
                if os.path.exists(scenario_landuse_geojson_path):
                    os.unlink(scenario_landuse_geojson_path)
            else:
                raise RunException('Exiting.  Use force option to overwrite.')

        if verbose:
            output = None
        else:
            output = open('/dev/null')
        try:
            # Connect to GI Notebook and fetch GI instance information (in GeoJSON format)
            # for this scenario ID.
            if verbose:
                self.outfp.write("\nDownloading GI scenario {0} from GI database...\n".format(scenario_id))
            nb = GINotebook(hostname=host,
                            api_root=api_root,
                            use_https=use_HTTPS, auth_token=auth_token)
            scenario = nb.get_scenario(scenario_id)
            scenario_geojson = scenario.get_instances_as_geojson(indent=2, shorten=True)
            (gi_scenario_data_wgs84, scenario_geojson_wgs84_path), (gi_scenario_data, scenario_geojson_path) = \
                self._write_geojson_and_reproject(scenario_geojson, gi_scenario_base, verbose=verbose, output=output)

            # Filter out instances that do not contain soils data
            gi_scenario_soils_base = "{0}_wsoils".format(gi_scenario_base)
            scenario_geojson_wsoils = scenario.get_instances_as_geojson(indent=2, shorten=True,
                                                                        filter=lambda a: a.get('e_1_pedid') is not None)
            (gi_scenario_soils_wgs84, scenario_soils_geojson_wgs84_path), (gi_scenario_soils, scenario_soils_geojson_path) = \
                self._write_geojson_and_reproject(scenario_geojson_wsoils, gi_scenario_soils_base,
                                                  verbose=verbose, output=output)

            # Import scenario GeoJSON into GRASS
            self._import_vector_into_grass(scenario_geojson_path, gi_scenario_data_key,
                                           force=force, verbose=verbose, output=output)

            # Import scenario (instances with soils data) GeoJSON into GRASS
            self._import_vector_into_grass(scenario_soils_geojson_path, gi_scenario_soils_data_key,
                                           force=force, verbose=verbose, output=output)

            # Generate raster layers from vector-based GI Scenario
            # Raster for updating soil type
            self._rasterize(gi_scenario_soils_data_key, gi_scenario_soils_data_key,
                            column='e_1_pedid', labelcolumn='e_1_pednm',
                            rast_title='GI soil types',
                            verbose=verbose,
                            force=force,
                            redir_fp=output)
            # Raster for updating stratum type
            gi_scenario_strata = "gi_scenario_strata"
            self._rasterize(gi_scenario_data_key, gi_scenario_strata,
                            column='e_1_vegid', labelcolumn='e_1_vegnm',
                            rast_title='GI vegetation types',
                            verbose=verbose,
                            force=force,
                            redir_fp=output)

            # Raster for updating land use
            # Filter out instances that are not rain gardens (i.e. the only GI type for which we currently have a
            #   land use).
            scenario_geojson_landuse = scenario.get_instances_as_geojson(indent=2, shorten=True,
                                                                         filter=lambda a: a.get('type',
                                                                                                '') == 'Rain Garden')
            (gi_scenario_landuse_wgs84, scenario_landuse_geojson_wgs84_path), \
            (gi_scenario_landuse, scenario_landuse_geojson_path) = \
                self._write_geojson_and_reproject(scenario_geojson_landuse, gi_scenario_landuse_base,
                                                  verbose=verbose, output=output)
            # Import land use (i.e. instances that are rain gardens) GeoJSON into GRASS
            self._import_vector_into_grass(scenario_landuse_geojson_path, gi_scenario_landuse_data_key,
                                           force=force, verbose=verbose, output=output)

            # Search for raster value for rain gardens in RHESSys parameter DB
            rg_name = 'raingarden'
            rg_found = self.param_db.search(self.param_const.SEARCH_TYPE_HIERARCHICAL, 'landuse', rg_name,
                                            None, None, None, None, None, None, None, None)
            if not rg_found:
                raise RunException("Unable to find raingarden landuse class in parameter database")

            rg_id = [c[1][2] for c in self.param_db.classes.iteritems()][0]

            # Generate raster layer from vector-based GI Scenario
            # Raster for updating landuse type
            self._rasterize_single_value(gi_scenario_landuse_data_key, gi_scenario_landuse_data_key,
                                         value=rg_id, label=rg_name,
                                         rast_title='GI landuse types',
                                         verbose=verbose,
                                         force=force,
                                         redir_fp=output)

            # Write out updated landuse, stratum, and soil rasters and parameter definitions
            # Backup landuse raster
            self._backup_raster(self.grassMetadata['landuse_rast'])
            # Update landuse raster
            self._update_raster(self.grassMetadata['landuse_rast'], gi_scenario_landuse_data_key)
            # Generate parameter definition file for landuse raster
            self._generate_parameter_definitions_for_raster(self.grassMetadata['landuse_rast'], 'landuse',
                                                            verbose=verbose)
            # Backup stratum raster
            self._backup_raster(self.grassMetadata['stratum_rast'])
            # Update stratum raster
            self._update_raster(self.grassMetadata['stratum_rast'], gi_scenario_strata)
            # Generate parameter definition file for stratum raster
            self._generate_parameter_definitions_for_raster(self.grassMetadata['stratum_rast'], 'stratum',
                                                            verbose=verbose)
            # Backup soils raster
            self._backup_raster(self.grassMetadata['soil_rast'])
            # Update soils raster
            self._update_raster(self.grassMetadata['soil_rast'], gi_scenario_soils_data_key)
            # Generate parameter definition file for soil raster
            self._generate_parameter_definitions_for_raster(self.grassMetadata['soil_rast'], 'soil',
                                                            verbose=verbose)
            # Write metadata
            RHESSysMetadata.writeGRASSEntry(self.context, "{0}_rast".format(gi_scenario_landuse_data_key),
                                            gi_scenario_landuse_data_key)
            RHESSysMetadata.writeGRASSEntry(self.context, "{0}_rast".format(gi_scenario_soils_data_key),
                                            gi_scenario_soils_data_key)
            RHESSysMetadata.writeGRASSEntry(self.context, "{0}_rast".format(gi_scenario_strata), gi_scenario_strata)

            RHESSysMetadata.writeGRASSEntry(self.context, "{0}_vect".format(gi_scenario_data_key), gi_scenario_data_key)
            RHESSysMetadata.writeGRASSEntry(self.context, "{0}_vect".format(gi_scenario_soils_data_key),
                                            gi_scenario_soils_data_key)
            RHESSysMetadata.writeGRASSEntry(self.context, "{0}_vect".format(gi_scenario_landuse_data_key),
                                            gi_scenario_landuse_data_key)
            RHESSysMetadata.writeRHESSysEntry(self.context, gi_scenario_data_key, gi_scenario_data)

            if verbose:
                self.outfp.write('\n\nFinished parameterizing GI.\n')

            # Write processing history
            RHESSysMetadata.appendProcessingHistoryItem(self.context, RHESSysMetadata.getCommandLine())

        finally:
            if output:
                output.close()

    def _rasterize_single_value(self, input, output, value, label, rast_title,
                                verbose=False, force=False, redir_fp=None):
        if verbose:
            self.outfp.write("\nRasterizing {title}...\n".format(title=rast_title))

        p = self.grassLib.script.start_command('v.to.rast',
                                               input=input,
                                               output=output,
                                               use='val',
                                               value=value,
                                               overwrite=force,
                                               quiet=not verbose,
                                               stdout=redir_fp,
                                               stderr=redir_fp)
        rc = p.wait()
        if rc != 0:
            raise RunException("Unable to rasterize {title}; v.to.rast returned {rc}".format(title=rast_title,
                                                                                             rc=rc))

        rule_file = tempfile.NamedTemporaryFile(delete=False)
        rule_file_name = rule_file.name
        rule_file.write("{value}:{label}\n".format(value=value, label=label))
        rule_file.close()

        p = self.grassLib.script.start_command('r.category',
                                               map=output,
                                               rules=rule_file_name)
        rc = p.wait()
        if rc != 0:
            raise RunException("Unable to assign category to raster {rast}; r.category returned {rc}".format(rast=output,
                                                                                                             rc=rc))
        os.unlink(rule_file_name)

    def _rasterize(self, input, output, column, labelcolumn, rast_title,
                   verbose=False, force=False, redir_fp=None):
        if verbose:
            self.outfp.write("\nRasterizing {title}...\n".format(title=rast_title))

        # Select only those GI instances that have information for column
        input_extract = "{input}_extract".format(input=input)
        p = self.grassLib.script.start_command('v.extract',
                                               input=input,
                                               output=input_extract,
                                               overwrite=True, # Temporary value, always overwrite
                                               quiet=not verbose,
                                               stdout=redir_fp,
                                               stderr=redir_fp)
        rc = p.wait()
        if rc != 0:
            raise RunException("Unable to extract features; v.extract returned {rc}".format(rc=rc))

        p = self.grassLib.script.start_command('v.to.rast',
                                               input=input_extract,
                                               output=output,
                                               column=column,
                                               labelcolumn=labelcolumn,
                                               overwrite=force,
                                               quiet=not verbose,
                                               stdout=redir_fp,
                                               stderr=redir_fp)
        rc = p.wait()
        if rc != 0:
            raise RunException("Unable to rasterize {title}; v.to.rast returned {rc}".format(title=rast_title,
                                                                                             rc=rc))
        # Remove extract vector
        p = self.grassLib.script.start_command('g.remove',
                                               flags='f',
                                               vect=input_extract,
                                               quiet=not verbose,
                                               stdout=redir_fp,
                                               stderr=redir_fp)
        rc = p.wait()
        if rc != 0:
            raise RunException("Unable to clean up {vect}; g.remove returned {rc}".format(vect=input_extract,
                                                                                          rc=rc))

    def _write_geojson_and_reproject(self, geojson, filename_base, verbose=False, output=None):
        # Write GeoJSON to file in project directory
        geojson_wgs84 = "{0}_wgs84.geojson".format(filename_base)
        geojson_wgs84_path = os.path.join(self.context.projectDir, geojson_wgs84)
        f = open(geojson_wgs84_path, 'w')
        f.writelines(geojson)
        f.close()

        # Reproject scenario data from WGS84 into our coordinate system (using ogr2ogr)
        geojson_reproj = "{0}.geojson".format(filename_base)
        geojson_reproj_path = os.path.join(self.context.projectDir, geojson_reproj)
        t_srs = self.studyArea['dem_srs']
        if verbose:
            self.outfp.write("\nReprojecting {file} to {srs}...\n".format(file=geojson_wgs84, srs=t_srs))
        path_to_ogr_cmd = self.context.config.get('GDAL/OGR', 'PATH_OF_OGR2OGR')
        ogr_cmd = "{ogr} -overwrite -f GeoJSON -t_srs {t_srs} {dest} {src}".format(ogr=path_to_ogr_cmd,
                                                                                   t_srs=t_srs,
                                                                                   dest=geojson_reproj_path,
                                                                                   src=geojson_wgs84_path)
        args = shlex.split(ogr_cmd)
        # Send output to /dev/null so that we don't see the spurious error that OGR
        # can't write GeoJSON files, when it seems to create ours just fine
        p = subprocess.Popen(args, stdout=output, stderr=output)
        rc = p.wait()
        if rc != 0:
            raise RunException("GI Instance re-project command {0} returned {1}".format(ogr_cmd,
                                                                                        rc))
        return (geojson_wgs84, geojson_wgs84_path), (geojson_reproj, geojson_reproj_path)

    def _import_vector_into_grass(self, vector_input, vector_output, force=False, verbose=False, output=None):
        if verbose:
                self.outfp.write('\nImporting GI scenario data into GRASS...\n')
        p = self.grassLib.script.start_command('v.in.ogr',
                                               dsn=vector_input,
                                               output=vector_output,
                                               overwrite=force,
                                               quiet=not verbose,
                                               stdout=output,
                                               stderr=output)
        rc = p.wait()
        if rc != 0:
            raise RunException("Unable to import {vect} into GRASS; v.in.ogr returned {rc}".format(vect=vector_input,
                                                                                     rc=rc))

    def _init_paramdb(self):
        param_db_path = os.path.join(self.context.projectDir, self.metadata['paramdb'])
        if not os.access(param_db_path, os.R_OK):
            sys.exit("Unable to read RHESSys parameters database {0}".format(param_db_path))
        param_db_path = os.path.abspath(param_db_path)

        sys.path.append( os.path.join(self.context.projectDir, self.metadata['paramdb_dir']) )
        params = importlib.import_module('rhessys.params')
        param_const = importlib.import_module('rhessys.constants')
        param_db = params.paramDB(filename=param_db_path)

        return (param_const, param_db)

    def _backup_raster(self, raster_name,
                       verbose=False, force=False, output=None):
        backup_name = "{0}_backup".format(raster_name)
        arg = "{0},{1}".format(raster_name, backup_name)

        if verbose:
                self.outfp.write("\nBacking up raster {rast} to {backup}\n".format(rast=raster_name,
                                                                                   backup_name=backup_name))
        p = self.grassLib.script.start_command('g.copy',
                                               rast=arg,
                                               overwrite=True,
                                               quiet=not verbose,
                                               stdout=output,
                                               stderr=output)
        rc = p.wait()
        if rc != 0:
            raise RunException("Unable to backup {rast}; g.copy returned {rc}".format(rast=raster_name,
                                                                                      rc=rc))

    def _update_raster(self, dest_raster, src_raster,
                      verbose=False, force=False, output=None):
        # Read category values of dest_ and src_raster
        raster_vals_list = self._read_raster_categories((dest_raster, src_raster))
        raster_vals = {}
        for cat_vals in raster_vals_list:
            for key in cat_vals:
                raster_vals[key] = cat_vals[key]
        mapcalc_expr = '$dest=if(!isnull($src), $src, $dest)'
        # import pdb; pdb.set_trace()
        self.grassLib.script.raster.mapcalc(mapcalc_expr, dest=dest_raster, src=src_raster,
                                            verbose=verbose)
        self._update_raster_categories(dest_raster, raster_vals, verbose=verbose, output=output)

    def _read_raster_categories(self, rasters,
                                verbose=False, output=None):
        raster_vals_list = []
        for raster in rasters:
            raster_vals = {}
            raster_vals_list.append(raster_vals)
            pipe = self.grassLib.script.pipe_command('r.stats', flags='licn', input=raster)
            for line in pipe.stdout:
                (dn, cat, num) = line.strip().split()
                if cat != 'NULL':
                    raster_vals[cat] = int(dn)
            pipe.wait()
        return raster_vals_list

    def _update_raster_categories(self, raster_name, cat_dict,
                                  verbose=False, output=None):
        rules_str = '\n'.join(["{cat}:{label}".format(cat=cat_dict[k], label=k) for k in cat_dict])
        rules_str += '\n'
        rule_file = tempfile.NamedTemporaryFile(delete=False)
        rule_file_name = rule_file.name
        rule_file.write(rules_str)
        rule_file.close()
        p = self.grassLib.script.start_command('r.category', map=raster_name, rules=rule_file_name,
                                               verbose=verbose)
        rc = p.wait()
        os.unlink(rule_file_name)
        if rc != 0:
            raise RunException("Unable to update categories for raster {rast}; r.category returned {rc}".format(rast=raster_name,
                                                                                                                rc=rc))

    def _generate_parameter_definitions_for_raster(self, raster_name, raster_type_name,
                                                   verbose=False):
        pipe = self.grassLib.script.pipe_command('r.stats', flags='licn', input=raster_name)
        raster_vals = {}
        for line in pipe.stdout:
            (dn, cat, num) = line.strip().split()
            if cat != 'NULL':
                raster_vals[cat] = int(dn)
        pipe.wait()
        if verbose:
            self.outfp.write("Writing GI {0} definition files to {1}".format(raster_type_name,
                                                                             self.paths.RHESSYS_DEF))
        for key in raster_vals.keys():
            if verbose:
                self.outfp.write("\n{rast_type} '{cat}' has dn {dn}".format(rast_type=raster_type_name,
                                                                            cat=key, dn=raster_vals[key]))
            params_found = self.param_db.search(self.param_const.SEARCH_TYPE_HIERARCHICAL, None, key, None, None, None,
                                                None, None, None, None, None,
                                                defaultIdOverride=raster_vals[key])
            assert(params_found)
            self.param_db.writeParamFileForClass(self.paths.RHESSYS_DEF)


#!/usr/bin/env python
"""@package RHESSysPlot

@brief Tool for visualizing basin-scale RHESSys output.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2014, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>

"""
import sys
import argparse
import math

import numpy as np
import pandas as pd
import statsmodels.api as sm
import matplotlib.pyplot as plt
import matplotlib

from rhessysworkflows.rhessys import RHESSysOutput

PLOT_TYPE_STD = 'standard'
PLOT_TYPE_LOGY = 'logy'
PLOT_TYPE_CDF = 'cdf'
PLOT_TYPE_SCATTER = 'scatter'
PLOT_TYPE_SCATTER_LOG = 'scatter-log'
PLOT_TYPES = [PLOT_TYPE_STD, PLOT_TYPE_LOGY, PLOT_TYPE_CDF, PLOT_TYPE_SCATTER, PLOT_TYPE_SCATTER_LOG]
PLOT_DEFAULT = PLOT_TYPE_STD

LINE_TYPE_LINE = 'line'
LINE_TYPE_DASH = 'dash'
LINE_TYPE_DASH_DOT = 'dashdot'
LINE_TYPE_COLON = 'colon'
LINE_TYPE_DOT = 'dot'
LINE_TYPE_DICT = { LINE_TYPE_LINE: '-',
                   LINE_TYPE_DASH: '--',
                   LINE_TYPE_DASH_DOT: '-.',
                   LINE_TYPE_COLON: ':',
                   LINE_TYPE_DOT: '.' }
LINE_TYPES = [LINE_TYPE_LINE, LINE_TYPE_DASH, LINE_TYPE_DASH_DOT, LINE_TYPE_COLON, LINE_TYPE_DOT]
NUM_LINE_TYPES = len(LINE_TYPES)


def plotGraphScatter(args, obs, data, log=False, sizeX=1, sizeY=1, dpi=80):
    
    assert( len(data) == len(args.legend) == 2 )
    
    if args.scatterwidth:
        linewidth = args.scatterwidth
    else:
        linewidth = 1
    
    fig = plt.figure(figsize=(sizeX, sizeY), dpi=dpi, tight_layout=False)
    ax = fig.add_subplot(111)

    x = data[0]
    y = data[1]
    min_val = min( min(x), min(y) )
    max_val = max( max(x), max(y) )
    floor = math.floor(min_val)
    ceil = math.ceil(max_val)
    range = np.linspace(floor, ceil, 1000)
    
    if not log:
        ax.plot(data[0], data[1], '.', markersize=linewidth*6)
        ax.plot(range, range, 'k-', linewidth=linewidth)
    else:
        ax.loglog(data[0], data[1], '.', markersize=linewidth*6)
        ax.loglog(range, range, 'k-', linewidth=linewidth)
    
    # Fit line
    (m, b) = np.polyfit(x, y, 1)
    fit_y = m*range
    if not log:
        ax.plot(range, fit_y, '--', color='grey', linewidth=linewidth)
    else:
        ax.loglog(range, fit_y, '--', color='grey', linewidth=linewidth)
    
    ax.set_xlim(floor, ceil)
    ax.set_ylim(floor, ceil)
    
    # Plot annotations
    
    # Annotate fit line, making sure the annotation does not overlap 1:1 line
    if m <= 1:
        if not log:
            ax.text(0.85*ceil, 0.8*max(fit_y), "$y = %.2f x$" % (m,), 
                    fontsize=args.ticklabelfontsize)
        else:
            ax.text(0.3*ceil, 0.2*max(fit_y), "$y = %.2f x$" % (m,), 
                    fontsize=args.ticklabelfontsize)
    else:
        if not log:
            ax.text(0.65*ceil, 0.8*max(fit_y), "$y = %.2f x$" % (m,),
                    fontsize=args.ticklabelfontsize)
        else:
            ax.text(0.3*ceil, 0.2*max(fit_y), "$y = %.2f x$" % (m,), 
                    fontsize=args.ticklabelfontsize)
    
    if args.title:
        title = args.title
    else:
        title = "%s vs. %s" % (args.legend[0], args.legend[1])
    fig.suptitle(title, y=0.98, fontsize=args.titlefontsize, fontweight=args.fontweight)
    
    # X-axis
    ax.set_xlabel(args.legend[0], fontsize=args.axesfontsize, fontweight=args.fontweight)
    plt.setp(ax.get_xticklabels(), fontsize=args.ticklabelfontsize)
    
    # Y-axis
    ax.set_ylabel(args.legend[1], fontsize=args.axesfontsize, fontweight=args.fontweight)
    plt.setp(ax.get_yticklabels(), fontsize=args.ticklabelfontsize)


def plotGraph(args, obs, data, sizeX=1, sizeY=1, dpi=80):
    
    fig = plt.figure(figsize=(sizeX, sizeY), dpi=dpi, tight_layout=True)
    ax = fig.add_subplot(111)
    
    if args.plottype == PLOT_TYPE_STD or \
       args.plottype == PLOT_TYPE_LOGY:
        x = obs.index
    elif args.plottype == PLOT_TYPE_CDF:
        x = np.linspace(min_x, max_x, num=len(obs) )
    
    # Plot observed values
    # Standard or log plot
    obs_y = obs
    if args.plottype == PLOT_TYPE_CDF:
        obs_ecdf = sm.distributions.ECDF(obs)
        obs_y = obs_ecdf(x)
    obs_plt = None
    if not args.supressObs:
        (obs_plt,) = ax.plot(x, obs_y, linewidth=2.0, color='black')
        
    # Plot modeled values 
    data_plt = []
    for (i, d) in enumerate(data):
        # Standard or log plot
        mod_y = d
        if args.plottype == PLOT_TYPE_CDF:
            mod_ecdf = sm.distributions.ECDF(d)
            mod_y = mod_ecdf(x)
        
        # Plot (we could move this outside of the for loop)
        if args.linewidth:
            linewidth = args.linewidth[i]
        else:
            linewidth = 1.0
            
        if args.linestyle:
            linestyle = LINE_TYPE_DICT[ args.linestyle[i] ]
        else:
            # Rotate styles
            styleIdx = ( (i + 1) % NUM_LINE_TYPES ) - 1
            linestyle = LINE_TYPE_DICT[ LINE_TYPES[styleIdx] ]
            
        if args.color:
            (mod_plt,) = ax.plot(x, mod_y, linewidth=linewidth, linestyle=linestyle,
                                 color=args.color[i])
        else:
            (mod_plt,) = ax.plot(x, mod_y, linewidth=linewidth, linestyle=linestyle)
        
        data_plt.append(mod_plt)
    
    # Plot annotations
    columnName = args.column.capitalize()
    if args.title:
        title = args.title
    else:
        if args.plottype == PLOT_TYPE_STD:
            title = columnName
        elif args.plottype == PLOT_TYPE_LOGY:
            title = "log(%s)" % (columnName,)
        elif args.plottype == PLOT_TYPE_CDF:
            title = "Cummulative distribution - %s" % (columnName,) 
    fig.suptitle(title, y=0.99)

    # X-axis
    if args.plottype == PLOT_TYPE_STD or \
       args.plottype == PLOT_TYPE_LOGY:
        num_years = len(x) / 365
        if num_years > 4:
            if num_years > 10:
                ax.xaxis.set_major_locator(matplotlib.dates.YearLocator())
            else:
                ax.xaxis.set_major_locator(matplotlib.dates.MonthLocator(interval=3))
        else:
            ax.xaxis.set_major_locator(matplotlib.dates.MonthLocator())
        ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%b-%Y'))
        # Rotate
        plt.setp( ax.xaxis.get_majorticklabels(), rotation=45)
        plt.setp( ax.xaxis.get_majorticklabels(), fontsize='x-small')
    
    if args.plottype == PLOT_TYPE_CDF:
        ax.set_xlim(min_x, max_x)
        ax.set_xscale('log')
        if args.xlabel:
            ax.set_xlabel(args.xlabel)
        else:
            ax.set_xlabel( columnName )
    elif args.xlabel:
        ax.set_xlabel(args.xlabel)
    
    # Y-axis
    if args.plottype == PLOT_TYPE_LOGY:
        ax.set_yscale('log')
    
    if args.ylabel:
        ax.set_ylabel(args.ylabel)
    elif args.plottype != PLOT_TYPE_CDF:
        y_label = columnName
        if args.plottype == PLOT_TYPE_LOGY:
            y_label = "log( %s )" % (columnName,)
        ax.set_ylabel( y_label )
    
    if args.supressObs:
        legend_items = args.legend
    else:
        data_plt.insert(0, obs_plt)
        legend_items = ['Observed'] + args.legend
    
    # Plot secondary data (if specified)
    if args.secondaryData and \
       (args.plottype == PLOT_TYPE_STD or args.plottype == PLOT_TYPE_LOGY):
        sec_file = open(args.secondaryData, 'r')
        (sec_datetime, sec_data) = RHESSysOutput.readColumnFromFile(sec_file,
                                                                    args.secondaryColumn,
                                                                    startHour=0)
        sec_file.close()
        sec = pd.Series(sec_data, index=sec_datetime)
        # Align timeseries
        (sec_align, obs_align) = sec.align(obs, join='inner')
        # Plot
        ax2 = ax.twinx()
        if args.secondaryPlotType == 'line':
            (sec_plot,) = ax2.plot(x, sec_align)
        elif args.secondaryPlotType == 'bar':
            sec_plot = ax2.bar(x, sec_align, facecolor='blue', edgecolor='none', width=2.0)
        secondaryLabel = args.secondaryColumn.capitalize()
        if args.secondaryLabel:
            secondaryLabel = args.secondaryLabel
        ax2.invert_yaxis()
        ax2.set_ylabel(args.secondaryLabel)
    #ax.set_zorder(ax2.get_zorder()+1) # put ax in front of ax2
    #ax.patch.set_visible(False) # hide the 'canvas' 
    
    # Plot legend last
    num_cols = len(data)
    if not args.supressObs:
        num_cols += 1
    
    if args.plottype == PLOT_TYPE_CDF:
        fig.legend( data_plt, legend_items, 'lower center', fontsize='x-small', 
                    bbox_to_anchor=(0.5, -0.015), ncol=num_cols, frameon=False )
    else:
        fig.legend( data_plt, legend_items, 'lower center', fontsize='x-small', 
                    bbox_to_anchor=(0.5, -0.01), ncol=num_cols, frameon=False )

if __name__ == "__main__":
    # Handle command line options
    parser = argparse.ArgumentParser(description='Plot CDF of N datasets vs. observed data')
    parser.add_argument('-p', '--plottype', required=False, 
                        default=PLOT_DEFAULT, choices=PLOT_TYPES,
                        help='Type of plot')
    parser.add_argument('-f', '--outfileSuffix', required=False,
                        help='Suffix to append on to name part of file name (i.e. before extension)')
    parser.add_argument('-o', '--obs', required=True,
                        help='File containing observed data')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-d', '--data', nargs='+',
                       help='One or more RHESSys output data files')
    group.add_argument('-b', '--behavioralData', nargs='+',
                       help='One or more ensemble output files from RHESSys behavioral runs')
    parser.add_argument('--color', required=False, nargs='+',
                        help='Color of symbol to be applied to plots of data. Color must be expressed in form recognized by matplotlib.')
    parser.add_argument('--linewidth', required=False, nargs='+', type=float,
                        help='Width of lines to be applied to plots of data. Value must be float in units of points.')
    parser.add_argument('--linestyle', required=False, nargs='+', choices=LINE_TYPES,
                        help='Style of symbol to be applied to plots of data. Styles correspond to those of matplotlib.')
    parser.add_argument('-c', '--column', required=True,
                        help='Name of column to use from data files')
    parser.add_argument('-t', '--title', required=False,
                        help='Title of figure')
    parser.add_argument('-l', '--legend', required=True, nargs='+',
                        help='Legend item labels')
    parser.add_argument('-x', '--xlabel', required=False,
                        help='X-axis label')
    parser.add_argument('-y', '--ylabel', required=False,
                        help='Y-axis label')
    parser.add_argument('--titlefontsize', required=False, type=float,
                        default=12)
    parser.add_argument('--scatterwidth', required=False, type=float,
                        default=1, help='Width to use for lines and markers in scatter plots.  Markers size will be determine by multiplying scatterwidth by 6.')
    parser.add_argument('--fontweight', required=False, 
                        choices=['ultralight','light','normal','regular','book','medium','roman','semibold','demibold','demi','bold','heavy','extra bold','black'],
                        default='regular')
    parser.add_argument('--legendfontsize', required=False, type=float,
                        default=6)
    parser.add_argument('--axesfontsize', required=False, type=float,
                        default=12)
    parser.add_argument('--ticklabelfontsize', required=False, type=float,
                        default=12)
    parser.add_argument('--figureX', required=False, type=int, default=4,
                        help='The width of the plot, in inches')
    parser.add_argument('--figureY', required=False, type=int, default=3,
                        help='The height of the plot, in inches')
    parser.add_argument('--supressObs', required=False, action='store_true',
                        help='Do not plot observed data.  Observed data will still be used for aligning timeseries. Not applicable to scatter plot output.')
    parser.add_argument('--secondaryData', required=False,
                        help='A data file containing the varaible to plot on a secondary Y-axis')
    parser.add_argument('--secondaryPlotType', required=False, choices=['bar', 'line'], default='bar',
                        help='Type of plot to use for secondary data.')
    parser.add_argument('--secondaryColumn', required=False,
                        help='Name of column to use from secondary data file')
    parser.add_argument('--secondaryLabel', required=False,
                        help='Label to use for seconary Y-axis')
    args = parser.parse_args()
    
    if args.color:
        if len(args.color) != len(args.data):
            sys.exit('Number of colors must match number of data files')
    
    if args.linewidth:
        if min(args.linewidth) <= 0.0:
            sys.exit('All line widths must be > 0.0')
        if len(args.linewidth) != len(args.data):
            sys.exit('Number of line widths must match number of data files')
            
    if args.linestyle:
        if len(args.linestyle) != len(args.data):
            sys.exit('Number of line styles must match number of data files')
    
    if args.secondaryData and not args.secondaryColumn:
        sys.exit('A secondary data file was specified, but the secondary column to use was not')
    
    if args.data and ( len(args.data) != len(args.legend) ):
        sys.exit('Number of legend items must equal the number of data files')
    elif args.behavioralData and ( len(args.behavioralData) != len(args.legend) ):
        sys.exit('Number of legend items must equal the number of data files')

    # Open data and align to observed
    obs_align = None
    data = []
    max_x = min_x = 0
    
    if args.data:
        # Open observed data
        obs_file = open(args.obs, 'r')
        (obs_datetime, obs_data) = RHESSysOutput.readObservedDataFromFile(obs_file,
                                                                          readHour=False)
        obs_file.close()
        obs = pd.Series(obs_data, index=obs_datetime)
        
        for d in args.data:
            mod_file = open(d, 'r')
            (tmp_datetime, tmp_data) = RHESSysOutput.readColumnFromFile(mod_file, args.column, startHour=0)
            tmp_mod = pd.Series(tmp_data, index=tmp_datetime)
            # Align timeseries
            (mod_align, obs_align) = tmp_mod.align(obs, join='inner')
            tmp_max_x = max(mod_align.max(), obs_align.max())
            if tmp_max_x > max_x:
                max_x = tmp_max_x
            min_x = max(min_x, mod_align.min())
        
            mod_file.close()
            data.append( mod_align )
    elif args.behavioralData:
        
        # Open observed data (behavioral data has hour in it, so we need to read obs. data differently)
        obs_file = open(args.obs, 'r')
        (obs_datetime, obs_data) = RHESSysOutput.readObservedDataFromFile(obs_file,
                                                                          readHour=True)
        obs_file.close()
        obs = pd.Series(obs_data, index=obs_datetime)
        
        for b in args.behavioralData:
            tmp_mod = pd.read_csv(b, index_col=0, parse_dates=True)
            # Convert df to series
            tmp_mod = pd.Series(tmp_mod[args.column], index=tmp_mod.index)
            # Align timeseries
            (mod_align, obs_align) = tmp_mod.align(obs, join='inner')
            tmp_max_x = max(mod_align.max(), obs_align.max())
            if tmp_max_x > max_x:
                max_x = tmp_max_x
            min_x = max(min_x, mod_align.min())
        
            data.append( mod_align )

    if args.plottype == PLOT_TYPE_SCATTER:
        plotGraphScatter(args, obs_align, data, log=False, 
                         sizeX=args.figureX, sizeY=args.figureY)
    elif args.plottype == PLOT_TYPE_SCATTER_LOG:
        plotGraphScatter(args, obs_align, data, log=True, 
                         sizeX=args.figureX, sizeY=args.figureY)
    else:
        plotGraph(args, obs_align, data, 
                  sizeX=args.figureX, sizeY=args.figureY)

    # Output plot
    filename = args.plottype
    if args.outfileSuffix:
        filename += '_' + args.outfileSuffix
    plot_filename_png = "%s.png" % (filename,)
    plot_filename_pdf = "%s.pdf" % (filename,)
    plt.savefig(plot_filename_png)
    plt.savefig(plot_filename_pdf)



"""@package climateio

@brief Routines for reading, modifying, and writing RHESSys climate data.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor 
      the names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>

"""
import os, errno
import re
import datetime

from rhessysworkflows.rhessys import readParameterFile


def getStartAndEndDateForClimateStation(climateStation, paths):
    """ Determine start and end date for the climate time-series associated
        with a climate station.  Supports only point time-series.  Does so
        by reading in the daily rain climate file.
    
        @param climateStation String representing path to climate station file
        @param paths rhessysworkflows.rhessys.RHESSysPaths
        
        @return Tuple of datetime.datetime representing start and end date of the climate
        time series; None is returned if the time-series lacks a start date
        
        @raise IOError if unable to read climate station file or daily rain 
        time-series
    """
    if not os.access(climateStation, os.R_OK):
        raise IOError("Unable to read climate station %s" % (climateStation,), errno.EACCES)
    
    station = readParameterFile(climateStation)
    
    rainTimeseriesPath = "%s.rain" % (os.path.join(paths.RHESSYS_DIR, station['daily_climate_prefix']), )
    if not os.access(rainTimeseriesPath, os.R_OK):
        raise IOError("Unable to read daily rain time-series %s" % (rainTimeseriesPath,), errno.EACCES)
    
    # Read daily rain timeseries to determine start and end date
    startDate = None
    endDate = None
    numDays = 0
    startDateRe = re.compile('^(?P<year>\d+)\s(?P<month>\d+)\s(?P<day>\d+)\s(?P<hour>\d+)\s*$')
    countDays = False
    
    f = open(rainTimeseriesPath, 'r')
    for line in f:
        if not countDays:
            line = line.strip()
            result = startDateRe.match(line)
            if result:
                startDate = datetime.datetime(year=int( result.group('year') ),
                                              month=int( result.group('month') ),
                                              day=int( result.group('day') ),
                                              hour=int( result.group('hour') ) )
                countDays = True
                continue
        # Count days
        numDays += 1
    
    if startDate:
        timeDelta = datetime.timedelta(days=numDays)
        endDate = startDate + timeDelta
    
    return (startDate, endDate)
    
    
    


"""@package rhessysworkflows.command.worldfile
    
@brief RHESSysWorkflows commands related to world files

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2015, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os
import sys
from subprocess import *

from rhessysworkflows.command.base import GrassCommand
from rhessysworkflows.command.exceptions import MetadataException
from rhessysworkflows.command.exceptions import RunException

from rhessysworkflows.rhessys import RHESSysPaths
from rhessysworkflows.metadata import RHESSysMetadata

class WorldfileMultiple(GrassCommand):
    
    def __init__(self, projectDir, configFile=None, outfp=sys.stdout):
        """ Construct a WorldfileMultiple command.
        Arguments:
        projectDir -- string    The path to the project directory
        configFile -- string    The path to an EcohydroLib configuration file
        outfp -- file-like object    Where output should be written to
        
        """
        super(WorldfileMultiple, self).__init__(projectDir, configFile, outfp)
    
    def checkMetadata(self):
        """ Check to make sure the project directory has the necessary metadata to run this command.
        """
        super(WorldfileMultiple, self).checkMetadata()
        
        # Check for necessary information in metadata
        if not 'basin_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a basin raster in a GRASS mapset" % (self.context.projectDir,))
        if not 'subbasins_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a sub-basin raster in a GRASS mapset" % (self.context.projectDir,))
        if not 'dem_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a DEM raster in a GRASS mapset" % (self.context.projectDir,)) 
        if not 'soil_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a soil raster in a GRASS mapset" % (self.context.projectDir,))
        if not 'patch_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a patch raster in a GRASS mapset" % (self.context.projectDir,))
        
        if not 'rhessys_dir' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a RHESSys directory" % (self.context.projectDir,))
        if not 'g2w_bin' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a grass2world executable" % (self.context.projectDir,))
        if not 'rat_bin' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain an AverageTables executable" % (self.context.projectDir,))
        if not 'template' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a world template" % (self.context.projectDir,))
        if not 'rhessys_dir' in self.metadata:
            raise MetadataException("Metadata in project directory {0} does not contain a RHESSys directory".format(self.context.projectDir))
            
    
    def run(self, *args, **kwargs):
        """ Multiple worldfiles, one worldfile for each subbasin delineated. 
        
        Arguments:
        verbose -- boolean    Produce verbose output. Default: False.
        """
        verbose = kwargs.get('verbose', False)
        
        self.checkMetadata()
        
        rhessysDir = self.metadata['rhessys_dir']
        self.paths = RHESSysPaths(self.context.projectDir, rhessysDir)
        
        templateFilename = os.path.basename(self.metadata['template'])
        templateFilepath = os.path.join(self.context.projectDir, self.metadata['template'])
        
        g2wPath = os.path.join(self.context.projectDir, self.metadata['g2w_bin'])
        
        # Make sure g2w can find rat
        g2wEnv = dict(os.environ)
        g2wEnv['PATH'] = self.paths.RHESSYS_BIN + os.pathsep + g2wEnv['PATH']
        
        # Make sure region is properly set
        demRast = self.grassMetadata['dem_rast']
        result = self.grassLib.script.run_command('g.region', rast=demRast)
        if result != 0:
            raise RunException("g.region failed to set region to DEM, returning {0}".format(result))
        
        # Mask subbasin to basin
        basin_rast = self.grassMetadata['basin_rast']
        result = self.grassLib.script.run_command('r.mask', flags='o', input=basin_rast, maskcats='1',
                                                  quiet=True)
        if result != 0:
            sys.exit("r.mask failed to set mask to basin, returning %s" % (result,))
        subbasin_raster = self.grassMetadata['subbasins_rast']
        subbasin_mask = "{0}_mask".format(subbasin_raster)
        mapcalc_input = "{subbasin_mask}={subbasins}".format(subbasin_mask=subbasin_mask,
                                                             subbasins=subbasin_raster)
        result = self.grassLib.script.write_command('r.mapcalc',
                                                    stdin=mapcalc_input,
                                                    stdout=PIPE,
                                                    stderr=PIPE)
        if result != 0:
            raise RunException("r.mapcalc failed to generate masked subbasin map {0}, input: {1}".format(subbasin_raster,
                                                                                                         mapcalc_input))
        
        # Get list of subbasins
        result = self.grassLib.script.read_command('r.stats', flags='n', input=subbasin_raster, quiet=True)
        if result is None or result == '':
            raise RunException("Error reading subbasin map {0}".format(subbasin_raster))
             
        subbasins = result.split()
        subbasin_masks = []
        worldfiles = []
        for subbasin in subbasins:
            # Remove mask
            result = self.grassLib.script.run_command('r.mask', flags='r', quiet=True)
            if result != 0:
                raise RunException("r.mask failed to remove mask")
            
            # Make a mask layer for the sub-basin
            mask_name = "subbasin_{0}".format(subbasin)
            subbasin_masks.append(mask_name)
            result = self.grassLib.script.write_command('r.mapcalc',
                                                        stdin="{mask_name}={subbasins} == {subbasin_number}".format(mask_name=mask_name,
                                                                                                                    subbasins=subbasin_mask,
                                                                                                                    subbasin_number=subbasin),
                                                        stdout=PIPE,
                                                        stderr=PIPE)
            if result != 0:
                raise RunException("r.mapcalc failed to generate mask for subbasin {0}".format(subbasin))
        
            # Mask to the sub-basin
            result = self.grassLib.script.run_command('r.mask', flags='o', input=mask_name, maskcats='1',
                                                      quiet=True)
            if result != 0:
                raise RunException("r.mask failed to set mask to sub-basin {0}, returning {1}".format(mask_name,
                                                                                                      result))
         
            worldfileName = "world_subbasin_{0}_init".format(subbasin)
            worldfilePath = os.path.join(self.paths.RHESSYS_WORLD, worldfileName)
            worldfiles.append(worldfilePath)
            g2wCommand = "{g2w} -t {template} -w {worldfile}".format(g2w=g2wPath, 
                                                                     template=templateFilepath, 
                                                                     worldfile=worldfilePath)
            
            if verbose:
                self.outfp.write("{0}\n".format(g2wCommand))
                self.outfp.write("\nRunning grass2world from {0}...".format(self.paths.RHESSYS_BIN))
                self.outfp.flush()

            cmdArgs = g2wCommand.split()
            process = Popen(cmdArgs, cwd=self.paths.RHESSYS_BIN, env=g2wEnv, 
                            stdout=PIPE, stderr=PIPE)
            (process_stdout, process_stderr) = process.communicate()
            if process.returncode != 0:
                raise RunException("grass2world failed, returning {0}".format(process.returncode))
     
            if verbose:
                self.outfp.write(process_stdout)
                self.outfp.write(process_stderr)
         
        # Remove mask
        result = self.grassLib.script.run_command('r.mask', flags='r', quiet=True)
        if result != 0:
            raise RunException("r.mask failed to remove mask") 
         
        # Write metadata
        RHESSysMetadata.writeRHESSysEntry(self.context, 'worldfiles_init', 
                                          RHESSysMetadata.VALUE_DELIM.join([self.paths.relpath(w) for w in worldfiles]))
        RHESSysMetadata.writeRHESSysEntry(self.context, 'subbasin_masks', 
                                          RHESSysMetadata.VALUE_DELIM.join([m for m in subbasin_masks]))

        if verbose:
            self.outfp.write('\n\nFinished creating worldfiles\n')

        # Write processing history
        RHESSysMetadata.appendProcessingHistoryItem(self.context, RHESSysMetadata.getCommandLine())



#!/usr/bin/env python
"""@package GeneratePatchMap

@brief Generate patch map in the GRASS location associated with the project directory.

@note Will set the zone map to be the patch map if no zone map exists

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'

2. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   dem_columns (if patch type is grid)

3. The following metadata entry(ies) must be present in the GRASS section of the metadata associated with the project directory:
   basin_rast
   dem_rast
   wetness_index_rast [if clump map wetness_index is selected]
   
4. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset
   
Post conditions
---------------
1. Will write the following entry(ies) to the GRASS section of metadata associated with the project directory:
   patch_rast
 
Usage:
@code
GeneratePatchMap.py -p /path/to/project_dir
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os, sys, shutil
import argparse

from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata

PATCH_RAST = 'patch'

# Handle command line options
parser = argparse.ArgumentParser(description='Generate patch maps or in GRASS location associated with the project directory.')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-t', '--patchType', dest='patchType', required=True, choices=['grid', 'clump'],
                    help='Type of patch to be generated: uniform grid or clumps based on elevation')
parser.add_argument('-c', '--clumpMap', dest='clumpMap', required=False, default='elevation', choices=['elevation', 'wetness_index'],
                    help='Type of patch to be generated: uniform grid or clumps based on elevation')
parser.add_argument('-z', '--forceZone', dest='forceZone', required=False, action='store_true',
                    help='Use patch map as zone map even if zone map is already defined. ' +
                    ' By default if a zone map is present, this script will not set the patch map as the zone map.')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing datasets in the GRASS mapset.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
studyArea = RHESSysMetadata.readStudyAreaEntries(context)
if not 'dem_rows' in studyArea:
    sys.exit("Metadata in project directory %s does not contain DEM rows entry" % (context.projectDir,))
    
grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'basin_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a basin raster" % (context.projectDir,))
if (args.patchType == 'clump') and (not 'dem_rast' in grassMetadata):
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a DEM raster" % (context.projectDir,))
if (args.clumpMap == 'wetness_index') and (not 'wetness_index_rast' in grassMetadata):
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a wetness index raster" % (context.projectDir,))


metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))

# Set up GRASS environment
modulePath = context.config.get('GRASS', 'MODULE_PATH')
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

# Make sure mask and region are properly set
demRast = grassMetadata['dem_rast']
result = grassLib.script.run_command('g.region', rast=demRast)
if result != 0:
    sys.exit("g.region failed to set region to DEM, returning %s" % (result,))

basinRast = grassMetadata['basin_rast']
result = grassLib.script.run_command('r.mask', flags='o', input=basinRast, maskcats='1')
if result != 0:
    sys.exit("r.mask failed to set mask to basin, returning %s" % (result,))

if args.patchType == 'grid':
    
    sys.stdout.write('Generating gridded patch map...\n')
    sys.stdout.flush()
    
    demCols = int(studyArea['dem_columns'])
    result = grassLib.script.write_command('r.mapcalc', 
                             stdin="%s=(row()-1) * %d + col()" % (PATCH_RAST, demCols) )
    if result != 0:
        sys.exit("r.mapcalc failed to create patch map, returning %s" % (result,))
    
if args.patchType == 'clump':
    if args.clumpMap == 'wetness_index':
        clumpMap = grassMetadata['wetness_index_rast']
    else:
        clumpMap = demRast
    
    sys.stdout.write("Generating clumped patch map based on %s raster...\n" % (clumpMap) )
    sys.stdout.flush()
    
    result = grassLib.script.run_command('r.clump', input=clumpMap, output=PATCH_RAST, overwrite=args.overwrite)
    if result != 0:
        sys.exit("r.mapcalc failed to create patch map, returning %s" % (result,))

sys.stdout.write('done\n')

# Write metadata    
RHESSysMetadata.writeGRASSEntry(context, 'patch_rast', PATCH_RAST)
if args.forceZone or not 'zone_rast' in grassMetadata:
    # Only overwrite zone raster
    RHESSysMetadata.writeGRASSEntry(context, 'zone_rast', PATCH_RAST)

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)


"""@package rhessysworkflows.command.flowtable
    
@brief RHESSysWorkflows commands related to flowtables

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2015, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os
import sys
from subprocess import *

from rhessysworkflows.command.base import GrassCommand
from rhessysworkflows.command.exceptions import MetadataException
from rhessysworkflows.command.exceptions import RunException

from rhessysworkflows.rhessys import RHESSysPaths
from rhessysworkflows.metadata import RHESSysMetadata

class FlowtableMultiple(GrassCommand):
    
    def __init__(self, projectDir, configFile=None, outfp=sys.stdout):
        """ Construct a FlowtableMultiple command.
        Arguments:
        projectDir -- string    The path to the project directory
        configFile -- string    The path to an EcohydroLib configuration file
        outfp -- file-like object    Where output should be written to
        
        """
        super(FlowtableMultiple, self).__init__(projectDir, configFile, outfp)
    
    def checkMetadata(self, *args, **kwargs):
        """ Check to make sure the project directory has the necessary metadata to run this command.
        """
        super(FlowtableMultiple, self).checkMetadata(args, kwargs)
        
        # Check for necessary information in metadata
        if not 'dem_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a DEM raster" % (self.context.projectDir,))
        if not 'slope_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a slope raster" % (self.context.projectDir,))
        if not 'streams_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a stream raster" % (self.context.projectDir,))
        if not 'zero_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a zero raster" % (self.context.projectDir,))
            
        if not 'dem_res_x' in self.studyArea:
            raise MetadataException("Metadata in project directory %s does not contain a DEM x resolution" % (self.context.projectDir,))
        if not 'dem_res_y' in self.studyArea:
            raise MetadataException("Metadata in project directory %s does not contain a DEM y resolution" % (self.context.projectDir,))
        
        if not 'rhessys_dir' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a RHESSys directory" % (self.self.context.projectDir,))
        if not 'cf_bin' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a createflowpaths executable" % (self.context.projectDir,))
        if not 'subbasin_masks' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain multiple worldfile masks" % (self.context.projectDir,))
        if not 'template' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a template" % (self.context.projectDir,))
        
        routeRoads = kwargs.get('routeRoads', False)
        if routeRoads:
            if not 'roads_rast' in self.grassMetadata:
                raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a roads raster" % (self.context.projectDir,))
        
        routeRoofs = kwargs.get('routeRoofs', False)
        if routeRoofs:
            if not 'roof_connectivity_rast' in self.grassMetadata:
                raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a roofs raster" % (self.context.projectDir,))
        if not 'impervious_rast' in self.grassMetadata:
                raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a impervious raster" % (self.context.projectDir,))
        
    def run(self, *args, **kwargs):
        """ Create flow tables for multiple worldfiles 
        
        Arguments:
        routeRoads -- boolean    Whether road routing should be enabled in createflowpaths. Default: False.
        routeRoofs -- boolean    Whether roof routing should be enabled in createflowpaths. Default: False.
        ignoreBurnedDEM -- boolean    If true, use the base DEM when running createflowpaths. 
                                      If false, use the stream-burned DEM (if present).  Default: False.
        force -- boolean        Whether to force createflowpaths to run if DEM X resolution != Y resolution. Default: False.
        verbose -- boolean    Produce verbose output. Default: False.
        """
        routeRoads = kwargs.get('routeRoads', False)
        routeRoofs = kwargs.get('routeRoofs', False)
        force = kwargs.get('force', False)
        ignoreBurnedDEM = kwargs.get('ignoreBurnedDEM', False)
        verbose = kwargs.get('verbose', False)
        
        self.checkMetadata(routeRoads=routeRoads, 
                           routeRoofs=routeRoofs)
        
        rhessysDir = self.metadata['rhessys_dir']
        self.paths = RHESSysPaths(self.context.projectDir, rhessysDir)
        
        demResX = float(self.studyArea['dem_res_x'])
        demResY = float(self.studyArea['dem_res_y'])
        if demResX != demResY:
            self.outfp.write("DEM x resolution (%f) does not match y resolution (%f)" %
                             (demResX, demResY) )
            if not force:
                raise RunException('Exiting.  Use force option to override')
        
        # Determine DEM raster to use
        demRast = self.grassMetadata['dem_rast']
        if ('stream_burned_dem_rast' in self.grassMetadata) and (not ignoreBurnedDEM):
            demRast = self.grassMetadata['stream_burned_dem_rast']
        self.outfp.write("Using raster named '%s' to calculate flow direction map\n" % (demRast,) )
        
        # Make sure region is properly set
        demRast = self.grassMetadata['dem_rast']
        result = self.grassLib.script.run_command('g.region', rast=demRast)
        if result != 0:
            raise RunException("g.region failed to set region to DEM, returning {0}".format(result))
        
        # Get paths for CF binary and template
        cfPath = os.path.join(self.context.projectDir, self.metadata['cf_bin'])
        templatePath = os.path.join(self.context.projectDir, self.metadata['template'])
        if verbose:
            self.outfp.write(self.templatePath)
            self.outfp.write('\n')
        
        # Make output file paths and file name templates
        flowTableNameBase = "world_{mask}"
        flowOutpath = os.path.join(self.paths.RHESSYS_FLOW, flowTableNameBase)
        cfOutpath = os.path.join(self.paths.RHESSYS_FLOW, "cf_{mask}.out")
        
        roads = None
        if routeRoads:
            roads = self.grassMetadata['roads_rast']
        else:
            roads = self.grassMetadata['zero_rast']
        
        roofs = None
        impervious = None
        
        # These filenames are only for metadata
        if routeRoofs:
            roofs = self.grassMetadata['roof_connectivity_rast']
            impervious = self.grassMetadata['impervious_rast']
            surfaceFlowtableTemplate = "world_{mask}_surface.flow"
            subsurfaceFlowtableTemplate = "world_{mask}_subsurface.flow"
        else:
            surfaceFlowtableTemplate = subsurfaceFlowtableTemplate = "world_{mask}.flow"
        
        # Make flowtable for each masked region
        if verbose:
            self.outfp.write('Running createflowpaths (this may take a few minutes)...')
            self.outfp.flush()
        
        surfaceFlowtables = []
        subsurfaceFlowtables = []
        masks = self.metadata['subbasin_masks'].split(RHESSysMetadata.VALUE_DELIM)
        for mask in masks:
            result = self.grassLib.script.run_command('r.mask', flags='o', input=mask, maskcats='1',
                                                      quiet=True)
            if result != 0:
                raise RunException("r.mask failed to set mask to sub-basin {0}, returning {1}".format(mask,
                                                                                                      result))
            # Run CF
            p = self.grassLib.script.pipe_command(cfPath, out=flowOutpath.format(mask=mask), 
                                                  template=templatePath, dem=demRast, 
                                                  slope=self.grassMetadata['slope_rast'],
                                                  stream=self.grassMetadata['streams_rast'],
                                                  road=roads, roof=roofs, impervious=impervious,
                                                  cellsize=demResX)
            (pStdout, pStderr) = p.communicate()
            
            if verbose:
                self.outfp.write("CF output:\n")
                self.outfp.write(pStdout)
                if pStderr:
                    self.outfp.write(pStderr)
            
            if p.returncode != 0:
                raise RunException("createflowpaths failed, returning %s" % ( str(p.returncode),))
            
            # Write cf output to project directory
            cfOut = open(cfOutpath.format(mask=mask), 'w')
            cfOut.write(pStdout)
            if pStderr:
                cfOut.write("\n\nStandard error output:\n\n")
                cfOut.write(pStderr)
            cfOut.close()
            
            surfFlow = os.path.join(self.paths.RHESSYS_FLOW, surfaceFlowtableTemplate.format(mask=mask))
            surfaceFlowtables.append(surfFlow)
            subsurfFlow = os.path.join(self.paths.RHESSYS_FLOW, subsurfaceFlowtableTemplate.format(mask=mask))
            subsurfaceFlowtables.append(subsurfFlow)
        
        # Remove mask
        result = self.grassLib.script.run_command('r.mask', flags='r', quiet=True)
        if result != 0:
            raise RunException("r.mask failed to remove mask") 
            
        # Write metadata
        cfCmd = "%s out=%s template=%s dem=%s slope=%s stream=%s road=%s roof=%s impervious=%s cellsize=%s" % \
        (cfPath, flowOutpath, templatePath, demRast, self.grassMetadata['slope_rast'],
         self.grassMetadata['streams_rast'], roads, roofs, impervious, demResX)
        RHESSysMetadata.writeRHESSysEntry(self.context, 'flowtable_cmd', cfCmd)
        RHESSysMetadata.writeRHESSysEntry(self.context, 'surface_flowtables', 
                                          RHESSysMetadata.VALUE_DELIM.join([self.paths.relpath(s) for s in surfaceFlowtables]) )
        RHESSysMetadata.writeRHESSysEntry(self.context, 'subsurface_flowtables', 
                                          RHESSysMetadata.VALUE_DELIM.join([self.paths.relpath(s) for s in subsurfaceFlowtables]) )

        if verbose:
            self.outfp.write('\n\nFinished creating flow tables\n')
            
        # Write processing history
        RHESSysMetadata.appendProcessingHistoryItem(self.context, RHESSysMetadata.getCommandLine())
         


#!/usr/bin/env python
"""@package GenerateLandcoverMaps

@brief Import landcover raster maps into a GRASS location and generate 
landcover, road and impervious coverage maps.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'

2. The following metadata entry(ies) must be present in the GRASS section of the metadata associated with the project directory:
   dem_rast
   landcover_rast
   stratum_rast [if defonly==False]
   landuse_rast [if defonly==False]
   
3. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   paramdb
   paramdb_dir
   grass_dbase
   grass_location
   grass_mapset
   rhessys_dir
   landcover_road_rule
   landcover_impervious_rule
   landcover_stratum_rule
   landcover_landuse_rule
   landcover_lai_rule [if LAI flag is specified]
   
Post conditions
---------------
1. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   stratum_defs
   landuse_defs
   
2. Will write the following entry(ies) to the GRASS section of metadata associated with the project directory:
   roads_rast [if skiproads == False] [if defonly==True]
   impervious_rast [if defonly==True]
   landuse_rast [if defonly==True]
   stratum_rast [if defonly==True]
   lai_rast [if LAI flag is specified] [if defonly==True]
 
Usage:
@code
GenerateLandcoverMaps.py -p /path/to/project_dir
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os, sys, shutil
import argparse
import importlib

from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

# Handle command line options
parser = argparse.ArgumentParser(description='Generate landcover maps in GRASS GIS')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-l', '--makeLaiMap', dest='makeLaiMap', required=False, action='store_true',
                    help='Make LAI map')
parser.add_argument('--skipRoads', dest='skipRoads', required=False, action='store_true', default=False,
                    help='Do not make roads map')
parser.add_argument('--defonly', dest='defonly', required=False, action='store_true',
                    help='Only generate landuse and stratum definition files, do not try to create maps.  Maps must already exist.')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing datasets in the GRASS mapset.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'landcover_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a landcover raster" % (context.projectDir,))
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a DEM raster in a GRASS mapset" % (context.projectDir,)) 

metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))
if not 'paramdb_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a ParamDB directory" % (context.projectDir,))
if not 'paramdb' in metadata:
    sys.exit("Metadata in project directory %s does not contain a ParamDB" % (context.projectDir,))

grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a DEM raster in a GRASS mapset" % (context.projectDir,)) 

paramDbPath = os.path.join(context.projectDir, metadata['paramdb'])
if not os.access(paramDbPath, os.R_OK):
    sys.exit("Unable to read RHESSys parameters database %s" % (paramDbPath,) )
paramDbPath = os.path.abspath(paramDbPath)

roadRulePath = os.path.join(context.projectDir, metadata['landcover_road_rule'])
if not os.access(roadRulePath, os.R_OK):
    sys.exit("Unable to read rule %s" % (roadRulePath,) )    
imperviousRulePath = os.path.join(context.projectDir, metadata['landcover_impervious_rule'])
if not os.access(imperviousRulePath, os.R_OK):
    sys.exit("Unable to read rule %s" % (imperviousRulePath,) )
landuseRulePath = os.path.join(context.projectDir, metadata['landcover_landuse_rule'])
if not os.access(landuseRulePath, os.R_OK):
    sys.exit("Unable to read rule %s" % (landuseRulePath,) )
stratumRulePath = os.path.join(context.projectDir, metadata['landcover_stratum_rule'])
if not os.access(stratumRulePath, os.R_OK):
    sys.exit("Unable to read rule %s" % (stratumRulePath,) )
laiRulePath = None
if args.makeLaiMap:
    laiRulePath = os.path.join(context.projectDir, metadata['landcover_lai_rule'])
    if not os.access(laiRulePath, os.R_OK):
        sys.exit("Unable to read rule %s" % (laiRulePath,) )

paths = RHESSysPaths(args.projectDir, metadata['rhessys_dir'])

# Import ParamDB from project directory
sys.path.append( os.path.join(context.projectDir, metadata['paramdb_dir']) )
params = importlib.import_module('rhessys.params')
paramConst = importlib.import_module('rhessys.constants')
paramDB = params.paramDB(filename=paramDbPath)

# Set up GRASS environment
modulePath = context.config.get('GRASS', 'MODULE_PATH')
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

landcoverRast = grassMetadata['landcover_rast']
demRast = grassMetadata['dem_rast']

# Make sure mask and region are properly set
result = grassLib.script.run_command('r.mask', flags='r')
if result != 0:
    sys.exit("r.mask filed, returning %s" % (result,) )
result = grassLib.script.run_command('g.region', rast=demRast)
if result != 0:
    sys.exit("g.region failed to set region to DEM, returning %s" % (result,))

# Reclassify landcover into stratum map
stratumRast = 'stratum'
if args.defonly:
    stratumRast = grassMetadata['stratum_rast']
else:
    # Create stratum map
    result = grassLib.script.read_command('r.reclass', input=landcoverRast, output=stratumRast, 
                               rules=stratumRulePath, overwrite=args.overwrite)
    if None == result:
        sys.exit("r.reclass failed to create stratum map, returning %s" % (result,))
    RHESSysMetadata.writeGRASSEntry(context, 'stratum_rast', stratumRast)

# Fetch relevant stratum default files from param DB
pipe = grassLib.script.pipe_command('r.stats', flags='licn', input=stratumRast)
rasterVals = {}
for line in pipe.stdout:
    (dn, cat, num) = line.strip().split()
    if cat != 'NULL':
        rasterVals[cat] = int(dn)
pipe.wait()
print("Writing stratum definition files to %s" % (paths.RHESSYS_DEF) )
for key in rasterVals.keys():
    print("stratum '%s' has dn %d" % (key, rasterVals[key]) )
    paramsFound = paramDB.search(paramConst.SEARCH_TYPE_HIERARCHICAL, None, key, None, None, None, None, None, None, None, None,
                                 defaultIdOverride=rasterVals[key])
    assert(paramsFound)
    paramDB.writeParamFileForClass(paths.RHESSYS_DEF)

# Reclassify landcover into landuse map
landuseRast = 'landuse'
if args.defonly:
    landuseRast = grassMetadata['landuse_rast']
else:
    result = grassLib.script.read_command('r.reclass', input=landcoverRast, output=landuseRast, 
                               rules=landuseRulePath, overwrite=args.overwrite)
    if None == result:
        sys.exit("r.reclass failed to create stratum map, returning %s" % (result,))
    RHESSysMetadata.writeGRASSEntry(context, 'landuse_rast', landuseRast)

# Fetch relevant landuse default files from param DB
pipe = grassLib.script.pipe_command('r.stats', flags='licn', input=landuseRast)
rasterVals = {}
for line in pipe.stdout:
    (dn, cat, num) = line.strip().split()
    if cat != 'NULL':
        rasterVals[cat] = int(dn)
pipe.wait()
print("Writing landuse definition files to %s" % (paths.RHESSYS_DEF) )
for key in rasterVals.keys():
    print("landuse '%s' has dn %d" % (key, rasterVals[key]) )
    paramsFound = paramDB.search(paramConst.SEARCH_TYPE_CONSTRAINED, None, key, None, None, None, None, None, None, None, None,
                                 limitToBaseClasses=True, defaultIdOverride=rasterVals[key])
    assert(paramsFound)
    paramDB.writeParamFileForClass(paths.RHESSYS_DEF)

# Reclassify landcover into road map
if not args.skipRoads and (not args.defonly):
    result = grassLib.script.read_command('r.reclass', input=landcoverRast, output='roads', 
                               rules=roadRulePath, overwrite=args.overwrite)
    if None == result:
        sys.exit("r.reclass failed to create roads map, returning %s" % (result,))
    RHESSysMetadata.writeGRASSEntry(context, 'roads_rast', 'roads')    

# Reclassify landcover into impervious map
if not args.defonly:
    result = grassLib.script.read_command('r.reclass', input=landcoverRast, output='impervious', 
                               rules=imperviousRulePath, overwrite=args.overwrite)
    if None == result:
        sys.exit("r.reclass failed to create impervious map, returning %s" % (result,))
    RHESSysMetadata.writeGRASSEntry(context, 'impervious_rast', 'impervious')    

# Reclassify landcover into LAI map
if args.makeLaiMap and (not args.defonly):
    if RHESSysMetadata.LC_RULE_LAI_COMPAT == os.path.basename(metadata['landcover_lai_rule']):
        # Compatibility mode, use r.reclass
        result = grassLib.script.read_command('r.reclass', input=landcoverRast, output='lai', 
                                              rules=laiRulePath, overwrite=args.overwrite)
        if None == result:
            sys.exit("r.reclass failed to create LAI map, returning %s" % (result,))
    else:
        # use r.reclass
        result = grassLib.script.read_command('r.recode', input=landcoverRast, output='lai', 
                                              rules=laiRulePath, overwrite=args.overwrite)
        if None == result:
            sys.exit("r.reclass failed to create LAI map, returning %s" % (result,))
        
    RHESSysMetadata.writeGRASSEntry(context, 'lai_rast', 'lai') 

# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'stratum_defs', True)
RHESSysMetadata.writeRHESSysEntry(context, 'landuse_defs', True)

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)


#!/usr/bin/env python
"""@package GenerateSoilTextureMap

@brief Import percent sand and percent clay raster maps into a GRASS location
and generate soil texture map using r.soils.texture. 

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'

2. The following metadata entry(ies) must be present in the manifest section of the metadata associated with the project directory:
   soil_raster_pctsand [if defonly==False]
   soil_raster_pctclay [if defonly==False]

3. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   paramdb
   paramdb_dir
   grass_dbase
   grass_location
   grass_mapset
   rhessys_dir
   
4. The following metadata entry(ies) must be present in the GRASS section of the metadata associated with the project directory:
   dem_rast
   soil_rast [if defonly==True]
   
5. Requires r.soils.texture GRASS extension: http://grasswiki.osgeo.org/wiki/GRASS_AddOns#r.soils.texture

Post conditions
---------------
1. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   soil_defs
   
2. Will write the following entry(ies) to the GRASS section of metadata associated with the project directory:
   soil_avgsand_rast [if defonly==False]
   soil_avgclay_rast [if defonly==False]
   soil_rast [if defonly==False]
   
Usage:
@code
GenerateSoilTextureMap.py -p /path/to/project_dir
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os, sys, errno
import argparse

from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

# Handle command line options
parser = argparse.ArgumentParser(description='Generate soil texture map for dataset in GRASS GIS')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('--defonly', dest='defonly', required=False, action='store_true',
                    help='Only generate soil definition files, do not try to create soil texture map.  Map must already exist.')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing datasets in the GRASS mapset.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

soil_rast = 'soil_texture'

# Check for necessary information in metadata
manifest = RHESSysMetadata.readManifestEntries(context)
if not args.defonly:
    if not 'soil_raster_pctsand' in manifest:
        sys.exit("Metadata in project directory %s does not contain a soil_raster_pctsand raster" % (context.projectDir,))
    if not 'soil_raster_pctclay' in manifest:
        sys.exit("Metadata in project directory %s does not contain a soil_raster_pctclay raster" % (context.projectDir,))

metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))
if not 'paramdb_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a ParamDB directory" % (context.projectDir,))
if not 'paramdb' in metadata:
    sys.exit("Metadata in project directory %s does not contain a ParamDB" % (context.projectDir,))

grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a DEM raster in a GRASS mapset" % (context.projectDir,)) 
demRast = grassMetadata['dem_rast']
if args.defonly:
    if not 'soil_rast' in grassMetadata:
        sys.exit("Metadata in project directory %s does not contain a soil raster in a GRASS mapset" % (context.projectDir,))
    soil_rast = grassMetadata['soil_rast'] 

paths = RHESSysPaths(args.projectDir, metadata['rhessys_dir'])

# Import ParamDB from project directory
paramDbPath = os.path.join(context.projectDir, metadata['paramdb'])
if not os.access(paramDbPath, os.R_OK):
    sys.exit("Unable to read RHESSys parameters database %s" % (paramDbPath,) )
sys.path.append( os.path.join(context.projectDir, metadata['paramdb_dir']) )
params = importlib.import_module('rhessys.params')
paramConst = importlib.import_module('rhessys.constants')
paramDB = params.paramDB(filename=paramDbPath)

# Set up GRASS environment
modulePath = context.config.get('GRASS', 'MODULE_PATH')
moduleEtc = context.config.get('GRASS', 'MODULE_ETC')
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

# Make sure mask and region are properly set
result = grassLib.script.run_command('r.mask', flags='r')
if result != 0:
    sys.exit("r.mask filed, returning %s" % (result,) )
result = grassLib.script.run_command('g.region', rast=demRast)
if result != 0:
    sys.exit("g.region failed to set region to DEM, returning %s" % (result,))

if not args.defonly:
    sys.stdout.write('Generating soil texture map from percent sand and clay maps...')
    sys.stdout.flush()
    # Import percent sand and percent clay raster maps into GRASS
    percentSandRasterPath = os.path.join(context.projectDir, manifest['soil_raster_pctsand'])
    result = grassLib.script.run_command('r.in.gdal', input=percentSandRasterPath, output='soil_pctsand', overwrite=args.overwrite)
    if result != 0:
        sys.exit("Failed to import soil_raster_pctsand into GRASS dataset %s/%s, results:\n%s" % \
                 (grassDbase, metadata['grass_location'], result) )
    RHESSysMetadata.writeGRASSEntry(context, 'soil_pctsand_rast', 'soil_pctsand')
        
    percentClayRasterPath = os.path.join(context.projectDir, manifest['soil_raster_pctclay'])
    result = grassLib.script.run_command('r.in.gdal', input=percentClayRasterPath, output='soil_pctclay', overwrite=args.overwrite)
    if result != 0:
        sys.exit("Failed to import soil_raster_pctclay into GRASS dataset %s/%s, results:\n%s" % \
                 (grassDbase, metadata['grass_location'], result) )
    RHESSysMetadata.writeGRASSEntry(context, 'soil_pctclay_rast', 'soil_pctclay')
    
    # Generate soil texture map
    schemePath = os.path.join(moduleEtc, 'USDA.dat')
    if not os.access(schemePath, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read r.soils.texture scheme %s" % (schemePath,) )
    soilTexture = os.path.join(modulePath, 'r.soils.texture')
    result = grassLib.script.read_command(soilTexture, sand='soil_pctsand', clay='soil_pctclay',
                                          scheme=schemePath, output=soil_rast, overwrite=args.overwrite)
    if None == result:
        sys.exit("r.soils.texture failed, returning %s" % (result,))
    RHESSysMetadata.writeGRASSEntry(context, 'soil_rast', soil_rast)
    
    sys.stdout.write('done\n')

sys.stdout.write('Reading soil parameter definitions from RHESSys ParamDB...')
sys.stdout.flush()

# Fetch relevant soil default files from param DB
pipe = grassLib.script.pipe_command('r.stats', flags='licn', input=soil_rast)
textures = {}
for line in pipe.stdout:
    (dn, cat, num, ) = line.strip().split()
    if cat != 'NULL':
        textures[cat] = int(dn)
pipe.wait()
print("Writing soil definition files to %s" % (paths.RHESSYS_DEF) )
for key in textures.keys():
    print("soil '%s' has dn %d" % (key, textures[key]) )
    paramsFound = paramDB.search(paramConst.SEARCH_TYPE_CONSTRAINED, None, key, None, None, None, None, None, None, None, None,
                                 limitToBaseClasses=True, defaultIdOverride=textures[key])
    assert(paramsFound)
    paramDB.writeParamFileForClass(paths.RHESSYS_DEF)

# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'soil_defs', True)

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)



"""@package rhessysworkflows.metadata
    
@brief Classes for writing and reading metadata for RHESSys workflows

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os
import errno
import ConfigParser

from ecohydrolib.metadata import GenericMetadata
import ecohydrolib.metadata as metadata
from ecohydrolib.metadata import MetadataVersionError

import ecohydrolib.command.landcover

import rhessysworkflows

class WorkflowVersionError(MetadataVersionError):
    def __init__(self, workflowVersion):
        self.workflowVersion = workflowVersion
        self._workflowVersion = rhessysworkflows.__version__
    def __str__(self):
        return repr("Version %s of RHESSysWorkflows was used to write the metadata, but you are running version %s." %\
                    (self.workflowVersion, self._workflowVersion) )


class RHESSysMetadata(GenericMetadata):
    """ Handles metadata persistance for RHESSys workflows.  Extends ecohydrolib.GenericMetadata class.
    
        @note All keys are stored in lower case.
        @note This object is stateless, all methods are static, writes to metadata store
        are written immediately.
    """
    
    _workflowVersion = rhessysworkflows.__version__
    
    VERSION_KEY = 'rhessysworkflows_version'
    
    MODEL_NAME = 'rhessys'
    RHESSYS_SECTION = MODEL_NAME
    # Patch list of valid sections
    GenericMetadata.SECTIONS.append(RHESSYS_SECTION)
   
    RULES_DIR = 'rules' 
    KNOWN_LC_TYPES = ecohydrolib.command.landcover.KNOWN_LC_TYPES
    LC_RULE_ROAD = 'road.rule'
    LC_RULE_IMPERVIOUS = 'impervious.rule'
    LC_RULE_LANDUSE = 'landuse.rule'
    LC_RULE_STRATUM = 'stratum.rule'
    LC_RULE_LAI_COMPAT = 'lai.rule'
    LC_RULE_LAI = 'lai-recode.rule'
    LC_RULES = [LC_RULE_ROAD, LC_RULE_IMPERVIOUS, LC_RULE_LANDUSE, LC_RULE_STRATUM, LC_RULE_LAI]
    SOILS_RULE = 'soils.rule'
    SOILS_RULES = [SOILS_RULE]
    
    @staticmethod
    def checkWorkflowVersion(projectDir):
        """ Check if metadata store is compatible with current version of RHESSysWorkflows. Accepts
            project directory as this method is used in the constructor to the Context class.
        
            @param projectDir, the path of the project whose metadata store is to be written to
            @raise WorkflowVersionError if a version already exists in the metadata store
            and is different than RHESSysMetadata._workflowVersion
        """
        metadataFilepath = os.path.join(projectDir, GenericMetadata.METADATA_FILENAME)
        if os.path.exists(metadataFilepath):
            if not os.access(metadataFilepath, os.R_OK):
                raise IOError(errno.EACCES, "Unable to read metadata store for project %s" % \
                              (projectDir,))
            # Read metadata store
            config = ConfigParser.RawConfigParser()
            config.read(metadataFilepath)
            if config.has_section(RHESSysMetadata.RHESSYS_SECTION):
                if config.has_option(RHESSysMetadata.RHESSYS_SECTION, \
                                 RHESSysMetadata.VERSION_KEY):
                    workflowVersion = config.get(RHESSysMetadata.RHESSYS_SECTION, \
                                     RHESSysMetadata.VERSION_KEY)
                    if workflowVersion != RHESSysMetadata._workflowVersion:
                        raise WorkflowVersionError(workflowVersion)


    @staticmethod
    def _writeWorkflowVersionToMetadata(config):
        """ Write RHESSysWorkflows version to RHESSYS_SECTION of metadata.
        
            @param config ConfigParser to write version information to
            @raise WorkflowVersionError if a version already exists in the metadata store
            and is different than RHESSysMetadata._workflowVersion
        """
        if not config.has_section(RHESSysMetadata.RHESSYS_SECTION):
            config.add_section(RHESSysMetadata.RHESSYS_SECTION)
        
        if not config.has_option(RHESSysMetadata.RHESSYS_SECTION, \
                             RHESSysMetadata.VERSION_KEY):
            config.set(RHESSysMetadata.RHESSYS_SECTION, \
                       RHESSysMetadata.VERSION_KEY, RHESSysMetadata._workflowVersion)
            return
            
        workflowVersion = config.get(RHESSysMetadata.RHESSYS_SECTION, \
                                     RHESSysMetadata.VERSION_KEY)
        if workflowVersion != RHESSysMetadata._workflowVersion:
            raise WorkflowVersionError(workflowVersion)
    
    @staticmethod
    def writeRHESSysEntry(context, key, value):
        """ Write a RHESSys entry to the metadata store for a given project.
            
            @note Will overwrite the value for a key that already exists
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param key The key to be written to the RHESSys section of the project metadata
            @param value The value to be written for key stored in the RHESSys section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.writeEntryToSection(context, RHESSysMetadata.RHESSYS_SECTION, key, value, \
                                            RHESSysMetadata._writeWorkflowVersionToMetadata)
    
    
    @staticmethod
    def deleteRHESSysEntry(context, key):
        """ Delete a RHESSys entry from the metadata store for a given project.
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be deleted from
            @param key The key to be deleted from the RHESSys section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.deleteEntryFromSection(context, RHESSysMetadata.RHESSYS_SECTION, key, \
                                               RHESSysMetadata._writeWorkflowVersionToMetadata)
         
     
    @staticmethod
    def readRHESSysEntries(context):
        """ Read all RHESSys entries from the metadata store for a given project
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            
            @exception A dictionary of key/value pairs from the RHESSys section of the project metadata
        """
        return GenericMetadata._readEntriesForSection(context.projectDir, RHESSysMetadata.RHESSYS_SECTION)
    

class ModelRun(metadata.ModelRun):
    # Register model name with EcohydroLib metadata
    GenericMetadata.MODEL_TYPES.append(RHESSysMetadata.MODEL_NAME)
    
    def __init__(self):
        super(ModelRun, self).__init__(RHESSysMetadata.MODEL_NAME)